import os
import logging

from nfixplanet.annotate import tools
from nfixplanet.annotate import processing
from nfixplanet import utils
from nfixplanet.constants import ANNOTATE_TOOLS

logger = logging.getLogger(__name__)


def run_annotate(
    input_genome: str,
    input_orf: str,
    input_hmm: str,
    output_dir: str,
    genomic_context_range: int,
    cpus: int,
):
    """Run nitrogen fixer annotation pipeline"""
    utils.check_external_tools(ANNOTATE_TOOLS)
    os.makedirs(output_dir, exist_ok=True)

    if input_genome:
        genome_path = input_genome
        orf_path = f"{output_dir}/prodigal_output.fna"
        hmm_path = f"{output_dir}/hmm_output.tbl"

        tools.prodigal(genome_path, orf_path)
        tools.hmmscan(orf_path, hmm_path, cpus)
    elif input_orf:
        orf_path = input_orf
        hmm_path = f"{output_dir}/hmm_output.tbl"
        logger.info("Input ORFs provided, skipping Prodigal")

        tools.hmmscan(orf_path, hmm_path, cpus)
    elif input_hmm:
        hmm_path = input_hmm
        logger.info("Input HMMs provided, skipping Prodigal and HMMscan")

    processing.filter_and_write_files(hmm_path, output_dir, genomic_context_range)
    logger.info("Pipeline completed")



from pathlib import Path
from dataclasses import dataclass
from importlib import resources

DATA_DIR = resources.files("nfixplanet.reference_data")

######################################
# Required by Annotate pipeline
######################################
HMM_PROFILE_PATH = str(DATA_DIR / "hmm_profiles" / "nfixplanet_models.hmm")


@dataclass
class GeneFamily:
    name: str
    required: dict[str, float]
    alternatives: list[dict[str, float]] | None
    optional: dict[str, float] | None


# NOTE: all values can be removed since we no longer do any
# manual filtering
GENE_FAMILIES = [
    GeneFamily(
        name="nif",
        required={"nifD": 583.6, "nifK": 460.0},
        alternatives=[{"nifH": 279.5, "vnfH": 150.6}],
        optional={"nifE": 504.6, "nifN": 530},
    ),
    GeneFamily(
        name="vnf",
        required={"vnfD": 933.0, "vnfK": 1151.5},
        alternatives=[{"nifH": 279.5, "vnfH": 150.6}],
        optional={},
    ),
    GeneFamily(
        name="anf",
        required={"anfD": 866.0, "anfK": 1016.1, "anfH": 626.6},
        alternatives=[],
        optional={},
    ),
    GeneFamily(
        name="chl",
        required={"ChlL": 194.0, "ChlB": 67.0, "ChlN": 23.0},
        alternatives=[],
        optional={},
    ),
    GeneFamily(
        name="nfl",
        required={"nflD": 88.0, "nflH": 509.0},
        alternatives=[],
        optional={},
    ),
]


ANNOTATE_TOOLS = [
    "prodigal",
    "hmmscan",
    "zcat",
]

######################################
# Required by Map pipeline
######################################

MAP_TOOLS = [
    "wget",
    "gunzip",
    "fastp",
    "hostile",
    "coverm",
]

# TODO: Make cache dir changable
HOSTILE_CACHE_DIR = Path("~/.local/share/hostile").expanduser()

NFIXPLANET_CACHE_DIR = Path("~/.local/share/nfixplanet").expanduser()
REFERENCE_FASTA_URL = (
    "https://zenodo.org/records/18772121/files/nfix_reference_v1.0.0.fna.gz?download=1"
)
REFERENCE_FASTA_NAME = NFIXPLANET_CACHE_DIR / "nfix_reference_v1.0.0.fna"
REFERENCE_INDEX_NAME = NFIXPLANET_CACHE_DIR / "nfix_reference_v1.0.0.fna.mmi"

MAPPING_REFERENCE = str(DATA_DIR / "map" / "mapping_ref_22.01.26.tsv")
TAXONOMY_REFERENCE = str(DATA_DIR / "map" / "taxonomy_ref_27.01.26.tsv")


TARGET_ANNOTATIONS = {
    "nifD",
    "nifK",
    "anfD",
    "anfK",
    "vnfD",
    "vnfK",
}



from src.nfixplanet.annotate.processing import filter_and_write_files

hmm_file = "/g/scb2/bork/ustick/nfixplanet_pipeline/01_annotation_pipeline/data/nfixplanet_test_full/02_test_hmm.tbl"

filter_and_write_files(hmm_file, "tmp_full", 10)


import gzip
import shutil
import logging
from pathlib import Path

from nfixplanet.profile import tools
from nfixplanet.profile import processing
from nfixplanet import utils
from nfixplanet.constants import (
    MAP_TOOLS,
    HOSTILE_CACHE_DIR,
    NFIXPLANET_CACHE_DIR,
    REFERENCE_INDEX_NAME,
)

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def ensure_dir(path: str):
    Path(path).mkdir(parents=True, exist_ok=True)


def run_profile(
    sample_id: str,
    r1: str | None,
    r2: str | None,
    single: str | None,
    output_dir: str,
    work_dir: str,
    cpus: int,
):
    """
    preprocess_fastqs -> clean_fastq -> coverm
    """
    utils.check_external_tools(MAP_TOOLS)
    ensure_dir(work_dir)

    fastp_dir = f"{work_dir}/processed_fastqs"
    clean_dir = f"{work_dir}/cleaned_fastqs"

    ensure_dir(fastp_dir)
    ensure_dir(clean_dir)

    processed_r1 = None
    processed_r2 = None
    processed_s = None

    # --------------------
    # FASTP
    # --------------------
    unpaired_file = None
    singles_file = None

    if r1 and r2:
        tools.check_files_exist([r1, r2])

        processed_r1 = f"{fastp_dir}/R1.fq.gz"
        processed_r2 = f"{fastp_dir}/R2.fq.gz"
        unpaired_file = f"{fastp_dir}/unpaired.fq"

        paired_html = f"{fastp_dir}/paired.html"
        paired_json = f"{fastp_dir}/paired.json"

        tools.fastp_paired(
            r1_in=r1,
            r2_in=r2,
            r1_out=processed_r1,
            r2_out=processed_r2,
            unpaired=unpaired_file,
            html=paired_html,
            json=paired_json,
            cpus=min(cpus, 4),
        )

    if single:
        tools.check_files_exist([single])

        singles_file = f"{fastp_dir}/singles.fq"

        single_html = f"{fastp_dir}/single.html"
        single_json = f"{fastp_dir}/single.json"

        tools.fastp_single(
            s_in=single,
            s_out=singles_file,
            html=single_html,
            json=single_json,
            cpus=cpus,
        )
        # --------------------
        # MERGE UNPAIRED + SINGLES
        # --------------------
        merged_rs = f"{fastp_dir}/RS.fq"
        merged_rs_gz = f"{fastp_dir}/RS.fq.gz"

        with open(merged_rs, "wb") as outfile:
            if singles_file and Path(singles_file).exists():
                with open(singles_file, "rb") as infile:
                    shutil.copyfileobj(infile, outfile)

            if unpaired_file and Path(unpaired_file).exists():
                with open(unpaired_file, "rb") as infile:
                    shutil.copyfileobj(infile, outfile)

        # gzip only if file has content
        if Path(merged_rs).exists() and Path(merged_rs).stat().st_size > 0:
            with open(merged_rs, "rb") as f_in, gzip.open(merged_rs_gz, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)

            processed_s = merged_rs_gz
        else:
            processed_s = None

    # --------------------
    # HOSTILE INDEX
    # --------------------
    hostile_cache_path = Path(HOSTILE_CACHE_DIR).expanduser()
    hostile_cache_path.mkdir(parents=True, exist_ok=True)
    if not any(hostile_cache_path.iterdir()):
        tools.hostile_index_fetch()

    # --------------------
    # HOSTILE CLEAN
    # --------------------
    cleaned_r1 = None
    cleaned_r2 = None
    cleaned_s = None

    if processed_r1 and processed_r2:
        cleaned_r1, cleaned_r2 = tools.hostile_clean_paired(
            r1=processed_r1,
            r2=processed_r2,
            out_dir=clean_dir,
            cpus=cpus,
        )

    if processed_s:
        cleaned_s = tools.hostile_clean_single(
            s=processed_s,
            out_dir=clean_dir,
            cpus=cpus,
        )

    # --------------------
    # MINIMAP INDEX
    # --------------------
    NFIXPLANET_CACHE_DIR.mkdir(parents=True, exist_ok=True)

    if not REFERENCE_INDEX_NAME.exists():
        tools.build_minimap_index(min(cpus, 3))

    # --------------------
    # COVERM
    # --------------------
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    coverage_file = f"{output_dir}/{sample_id}_coverage.tsv"

    tools.coverm_contig(
        r1=cleaned_r1,
        r2=cleaned_r2,
        single=cleaned_s,
        reference_index=str(REFERENCE_INDEX_NAME),
        output_file=coverage_file,
        cpus=cpus,
    )

    # --------------------
    # Process results
    # --------------------
    gene_file = f"{output_dir}/{sample_id}_gene_table.tsv"
    otu_file = f"{output_dir}/{sample_id}_OTU_table.tsv"
    processing.annotate_mapping_results(coverage_file, gene_file, otu_file)
    processing.profile_results(otu_file, output_dir)

    logger.info(f"Finished calculating coverage score for sample {sample_id}")





import re
import pandas as pd
from pathlib import Path

from nfixplanet.constants import (
    MAPPING_REFERENCE,
    TARGET_ANNOTATIONS,
    TAXONOMY_REFERENCE,
)


def load_otu_table(sample_path: str) -> pd.DataFrame:
    """Get means for contigs"""
    df = pd.read_csv(sample_path, sep="\t")

    # If paired and single
    if df.shape[1] == 5:
        mean_cols = df.columns[df.columns.str.contains("Mean")]
        row_sum = df[mean_cols].sum(axis=1)
        df["Mean"] = row_sum
    # Just rename columns
    else:
        df.columns = ["Contig", "Covered_Bases", "Mean"]

    # only return mean
    return df[["Contig", "Mean"]]


def extract_sample_name(path: str) -> str:
    match = re.search(r"(.*?)_coverage\.tsv", Path(path).name)
    if not match:
        raise ValueError("Could not extract sample name")
    return match.group(1)


def compute_target_otu_table(merged_df: pd.DataFrame, sample_name: str) -> pd.DataFrame:
    """Filter merged data to target annotations and rename Mean column to sample name."""
    subset = merged_df[merged_df["annotation"].isin(TARGET_ANNOTATIONS)]
    otu_table = subset[["map_id", "annotation", "Mean"]].copy()
    otu_table.rename(columns={"Mean": sample_name}, inplace=True)
    return otu_table


def compute_genome_normalized_table(
    reference_df: pd.DataFrame, otu_df: pd.DataFrame
) -> pd.DataFrame:
    """
    Compute genome-level normalized abundances by averaging values across genes per genome_contig.
    """

    # attach genome_contig
    contig_map = reference_df[["map_id", "genome_contig"]]
    merged = contig_map.merge(otu_df, on="map_id", how="inner")

    # sum per genome_contig
    grouped = (
        merged.drop(columns=["map_id", "annotation"])
        .groupby("genome_contig", as_index=False)
        .sum()
    )

    # count genes per genome
    gene_counts = merged.groupby("genome_contig").size().reset_index(name="Freq")

    normalized = grouped.merge(gene_counts, on="genome_contig")

    sample_cols = normalized.columns.difference(["genome_contig", "Freq"])

    normalized[sample_cols] = normalized[sample_cols].div(normalized["Freq"], axis=0)

    normalized = normalized.drop(columns=["Freq"])

    # Sort output to match R
    normalized = normalized.sort_values(
        "genome_contig", key=lambda x: x.str.lower()
    ).reset_index(drop=True)

    return normalized


def annotate_mapping_results(coverage_file: str, gene_file: str, otu_file: str):
    """Process coverage data to generate gene-level and genome-normalized OTU tables."""
    reference_df = pd.read_csv(MAPPING_REFERENCE, sep="\t")
    otu_df = load_otu_table(coverage_file)

    sample_name = extract_sample_name(coverage_file)

    # merge with annotation
    merged_df = reference_df.merge(
        otu_df, left_on="map_id", right_on="Contig", how="inner"
    )

    # Aggregate Mean coverage values by annotation and label column with sample name
    gene_result = (
        merged_df.groupby("annotation", as_index=False)["Mean"]
        .sum()
        .rename(columns={"Mean": sample_name})
    )

    otu_table = compute_target_otu_table(merged_df, sample_name)
    otu_normalized = compute_genome_normalized_table(reference_df, otu_table)

    gene_result.to_csv(gene_file, sep="\t", index=False, float_format="%.15g")

    otu_normalized.to_csv(otu_file, sep="\t", index=False, float_format="%.15g")


def sum_otu_by_group(
    group_col: str,
    otu_df: pd.DataFrame,
    tax_df: pd.DataFrame,
    output_dir: str,
    out_prefix: str = "OTU_group_summed",
):
    """Sum OTU abundance values grouped by a specified taxonomy column and write to file."""
    tax_map = tax_df[["genome_contig", group_col]]
    merged = otu_df.merge(tax_map, on="genome_contig", how="inner")

    numeric_cols = merged.select_dtypes(include="number").columns.tolist()

    grouped = merged.groupby(group_col, dropna=False)[numeric_cols].sum().reset_index()

    out_file = f"{output_dir}/{out_prefix}_by_{group_col}.tsv"
    grouped.to_csv(out_file, sep="\t", index=False, float_format="%.15g")


def profile_results(otu_path: str, output_dir: str):
    """Generate taxonomy-level OTU summaries for standard taxonomic ranks."""
    # Stop autoconverting "" to "NaN"
    tax_df = pd.read_csv(TAXONOMY_REFERENCE, sep="\t", keep_default_na=False)
    otu_df = pd.read_csv(otu_path, sep="\t")

    for rank in ["d", "p", "c", "o", "f", "g", "s"]:
        sum_otu_by_group(rank, otu_df, tax_df, output_dir)



import subprocess
import logging
import sys

from nfixplanet.constants import HMM_PROFILE_PATH

logger = logging.getLogger(__name__)


def prodigal(input_file: str, out_file: str):
    """Run prodigal

    Arguments:
        input_file (str): full path of input fasta
        out_file (str): fullpath of output file
    """
    logger.info("Running prodigal...")
    try:
        if input_file.endswith(".gz"):
            uncompressed = subprocess.Popen(
                ("zcat", input_file), stdout=subprocess.PIPE
            )
            subprocess.check_output(
                [
                    "prodigal",
                    "-i",
                    "/dev/stdin",
                    "-d",
                    out_file,
                    "-o",
                    "/dev/null",
                    "-p",
                    "meta",
                    "-q",
                ],
                stdin=uncompressed.stdout,
                universal_newlines=True,
            )
        else:
            subprocess.check_output(
                [
                    "prodigal",
                    "-i",
                    input_file,
                    "-d",
                    out_file,
                    "-o",
                    "/dev/null",
                    "-p",
                    "meta",
                    "-q",
                ],
                universal_newlines=True,
            )
        logger.info("Prodigal completed")
    except subprocess.CalledProcessError as e:
        logger.error(f"Failed to run Prodigal on {input_file}: {e}")
        sys.exit(1)


def hmmscan(input_file: str, out_file: str, cpus: int):
    """Run hmmscan

    Arguments:
        input_file (str): full path of input fasta (fna)
        out_file (str): fullpath of output file
    """
    logger.info("Running hmmscan...")
    try:
        subprocess.check_output(
            [
                "hmmscan",
                "--cpu",
                str(cpus),
                "--cut_ga",
                "--tblout",
                out_file,
                HMM_PROFILE_PATH,
                input_file,
            ],
            universal_newlines=True,
        )
        logger.info("Hmmscan completed")
    except subprocess.CalledProcessError as e:
        logger.error(f"Failed to run hmmscan on {input_file}: {e}")
        sys.exit(1)



import logging
import pandas as pd

from nfixplanet.constants import GeneFamily, GENE_FAMILIES

logger = logging.getLogger(__name__)


def load_hmm_output(path: str) -> pd.DataFrame:
    column_names = [
        "target_name",
        "accession",
        "query_name",
        "full_accession",
        "full_evalue",
        "full_score",
        "full_bias",
        "domain_evalue",
        "domain_score",
        "domain_bias",
        "exp",
        "reg",
        "clu",
        "ov",
        "env",
        "dom",
        "rep",
        "inc",
    ]
    df = pd.read_csv(
        filepath_or_buffer=path,
        delim_whitespace=True,
        names=column_names,
        usecols=range(len(column_names)),
        comment="#",
    )  # type: ignore
    ordered_cols = ["query_name", "accession"] + [
        col for col in column_names if col not in ["query_name", "accession"]
    ]
    df = df[ordered_cols]
    # NOTE: contig and gene number are needed for neighborhood filtering
    # and will be removed before writing output
    df[["contig", "gene"]] = df["query_name"].str.rsplit("_", n=1, expand=True)
    df["gene"] = df["gene"].astype(int)
    return df


def get_best_hmm_hits(df: pd.DataFrame) -> pd.DataFrame:
    """
    For each query, keep only the best-scoring HMM match:
    lowest e-value wins; ties broken by highest bit score
    """
    return (
        df.sort_values(["full_evalue", "full_score"], ascending=[True, False])
        .groupby("query_name", sort=False)
        .first()
        .reset_index()
    )


def filter_hits_by_family(df: pd.DataFrame, gene_family: GeneFamily) -> pd.DataFrame:
    """
    Keep only rows where the target gene contains all required family members.
    """
    mask = pd.Series(False, index=df.index)

    for gene, _ in gene_family.required.items():
        mask |= (df["target_name"] == gene)

    if gene_family.alternatives:
        for group in gene_family.alternatives:
            for gene, _ in group.items():
                mask |= (df["target_name"] == gene)

    return df[mask]


def gene_family_to_dict(gf: GeneFamily) -> dict[str, list[int]]:
    result = {}

    result.update({k: [] for k in gf.required})

    if gf.alternatives:
        for alt_dict in gf.alternatives:
            result.update({k: [] for k in alt_dict})

    if gf.optional:
        result.update({k: [] for k in gf.optional})

    return result


def is_within_genomic_context_range(positions: pd.Series, genomic_context_range: int):
    return (positions.max() - positions.min()) <= genomic_context_range


def is_optional_gene_within_genomic_context_range(
    optional_gene_pos: int, positions: pd.Series, genomic_context_range: int
):
    # TODO: ask lucas if optional gene can be min - genomic_context_range
    return abs(optional_gene_pos - positions.max()) <= genomic_context_range  # or (
    #     abs(optional_gene_pos - positions.min()) <= genomic_context_range
    # )


def resolve_gene_key(
    target_name: str, gene_family: GeneFamily, alt_genes: set[str] | None
) -> str:
    """Convert both nifH and vnfH to nif/vnf depending on gene_family"""
    if gene_family.name == "nif" and alt_genes and target_name in alt_genes:
        return "nifH"
    elif gene_family.name == "vnf" and alt_genes and target_name in alt_genes:
        return "vnfH"
    else:
        return target_name


def get_valid_contig_indexes(
    hmm_df: pd.DataFrame, gene_family: GeneFamily, genomic_context_range: int
):
    """
    Get indices for rows that meet the following criteria:
    1. all required genes are present
    2. at least one alternative gene is present
    3. all genes in the above set are withint the genomic_context_range
    4. Optional genes are within the genomic_context_range of the largest value
    """
    genes_to_indices = gene_family_to_dict(gene_family)

    for contig, group in hmm_df.groupby("contig"):
        genes_present = set(group["target_name"])
        # NOTE: this will break if there are more than one set of alternatives
        # Fix if that happens
        alt_genes = (
            set(gene_family.alternatives[0].keys())
            if gene_family.alternatives
            else None
        )
        optional_genes = (
            set(gene_family.optional.keys()) if gene_family.optional else set()
        )

        # All required genes not present
        if not all(g in genes_present for g in gene_family.required):
            continue

        # at least one alternate gene is present
        if alt_genes and not any(g in genes_present for g in alt_genes):
            continue

        if group["target_name"].duplicated().any():
            # duplicates = group[group["target_name"].duplicated(keep=False)]
            logger.info(f"Duplicates found for {contig}")

        # Sort by gene position to enable sliding window
        group_sorted = group.sort_values("gene")
        positions = group_sorted["gene"].values
        genes = group_sorted["target_name"].values
        indices = group_sorted.index.values

        # Slide a window of size genomic_context_range over all rows,
        # checking if any window contains all required genes (+ alt requirement)
        for i in range(len(group_sorted)):
            # TODO: figure out early exit for speedup

            # select widow of position + genomic_context_range
            window_mask = (positions >= positions[i]) & (
                positions <= positions[i] + genomic_context_range
            )
            window_names = set(genes[window_mask])
            window_indices = indices[window_mask]
            window_rows = group_sorted.loc[window_indices]

            # All required genes must be present in this window
            if not all(g in window_names for g in gene_family.required):
                continue

            # At least one alternate gene must be present (if alternatives defined)
            if alt_genes and not any(g in window_names for g in alt_genes):
                continue

            # All rows in this valid window are valid so add their indices
            for idx, row in window_rows.iterrows():
                # TODO: figure out early exit for speedup
                gene_key = resolve_gene_key(row["target_name"], gene_family, alt_genes)

                if gene_key in optional_genes:
                    if not is_optional_gene_within_genomic_context_range(
                        row["gene"],
                        group_sorted.loc[window_indices, "gene"],
                        genomic_context_range,
                    ):
                        continue

                if idx not in genes_to_indices[gene_key]:
                    genes_to_indices[gene_key].append(idx)  # type: ignore

    return genes_to_indices


def get_genes_to_df(
    genes_to_indexes: dict[str, list[int]],
    hmm_df: pd.DataFrame,
    gene_family: GeneFamily,
) -> dict[str, pd.DataFrame]:
    genes_to_df = {}
    for gene, indexes in genes_to_indexes.items():
        if not indexes:
            logger.info(f"No hits for {gene} ({gene_family.name})")
            continue
        # logger.debug(f"family: {gene_family.name}\tgene: {gene}")
        # logger.debug(hmm_df.head)
        genes_to_df[gene] = hmm_df.loc[indexes].drop(columns=["contig", "gene"])

    return genes_to_df


def get_filtered_top_hits(
    best_hmm_hits: pd.DataFrame,
    gene_family: GeneFamily,
    genomic_context_range: int,
) -> dict[str, pd.DataFrame] | None:
    """
    Filter hits that meet the score threshold, gene requirements
    and neighborhood requirements
    """
    # Step 1: filter contigs to only keep genes from the family
    hits_filtered_by_family = filter_hits_by_family(best_hmm_hits, gene_family)
    if hits_filtered_by_family.empty:
        logger.info(f"No hits above threshold for {gene_family.name}")
        return None

    # Step 2: Get indexes which meet all filtering requirements
    genes_to_indexes = get_valid_contig_indexes(
        hits_filtered_by_family, gene_family, genomic_context_range
    )

    # Step 3: Convert indexes to df
    genes_to_df = get_genes_to_df(genes_to_indexes, best_hmm_hits, gene_family)

    return genes_to_df


def write_tsvs(genes_to_df: dict[str, pd.DataFrame], output_dir: str):
    for gene, df in genes_to_df.items():
        path = f"{output_dir}/{gene}.tsv"
        df.to_csv(path, sep="\t", index=False)
        logger.info(f"Created file: {path}")


def filter_and_write_files(path: str, output_dir: str, genomic_context_range: int):
    hmm_output = load_hmm_output(path)
    best_hmm_hits = get_best_hmm_hits(hmm_output)
    for family in GENE_FAMILIES:
        genes_to_df = get_filtered_top_hits(
            best_hmm_hits, family, genomic_context_range
        )
        if genes_to_df:
            write_tsvs(genes_to_df, output_dir)



import argparse
import logging
from importlib.metadata import version

from nfixplanet.annotate.pipeline import run_annotate
from nfixplanet.profile.pipeline import run_profile
from nfixplanet import utils


def get_version() -> str:
    return version("nfixplanet")


def get_parser():
    parser = argparse.ArgumentParser(
        description="A pipeline for the detection of nitrogen fixers."
    )

    parser.add_argument(
        "--version",
        action="version",
        help="Print version number and exit.",
        version=get_version(),
    )

    subparsers = parser.add_subparsers(dest="command", required=True)

    # -------------------
    # annotate subcommand
    # -------------------
    annotate_parser = subparsers.add_parser(
        "annotate",
        help="Run annotation pipeline",
    )

    annotate_parser.add_argument(
        "--input_genomes",
        help="Path to input genom fasta file (can be gzipped)",
        default="",
    )
    annotate_parser.add_argument(
        "--input_orfs",
        help="Path to input ORF fasta file (Prodigal output)",
        default="",
    )

    annotate_parser.add_argument(
        "--input_hmms",
        help="Path to HMM tables (HMMER output)",
        default="",
    )

    annotate_parser.add_argument(
        "--output_directory",
        required=True,
        help="Path to output directory",
    )

    annotate_parser.add_argument(
        "--genomic_context_range",
        type=int,
        default=10,
        help="Maximum number of genes upstream or downstream to consider for the operon context (default: 10)",
    )

    annotate_parser.add_argument(
        "--cpus",
        type=int,
        default=2,
        help="Number of CPUs to use for HMMscan (default: 2, max recommended: 4)",
    )

    annotate_parser.add_argument(
        "--verbose",
        action="store_true",
    )

    annotate_parser.set_defaults(func=run_annotate_command)

    # ---------------
    # map subcommand
    # ---------------
    profile_parser = subparsers.add_parser(
        "map",
        help="Run profiling pipeline",
    )

    profile_parser.add_argument(
        "--sample_id",
        type=str,
        required=True,
        help="Name of FASTA/FASTQ sample",
    )

    profile_parser.add_argument(
        "--read_1",
        type=str,
        help="Path to FASTA/FASTQ file for paired-end read 1 (R1).",
    )

    profile_parser.add_argument(
        "--read_2",
        type=str,
        help="Path to FASTA/FASTQ file for paired-end read 2 (R2). Must be provided with --read_1.",
    )

    profile_parser.add_argument(
        "--single",
        type=str,
        help="Path to FASTA/FASTQ file for single-end reads.",
    )

    profile_parser.add_argument(
        "--output_directory",
        type=str,
        required=True,
        help="Path to output directory",
    )

    profile_parser.add_argument(
        "--work_directory",
        type=str,
        default="tmp",
        help="Path to directory for temporary files",
    )

    profile_parser.add_argument(
        "--cpus",
        type=int,
        default=8,
        help="Number of CPUs used by processes",
    )

    profile_parser.add_argument(
        "--verbose",
        action="store_true",
    )

    profile_parser.set_defaults(func=run_profile_command)

    return parser


def run_annotate_command(args):
    logger = logging.getLogger(__name__)
    logger.debug(args)

    if not (args.input_genomes or args.input_orfs or args.input_hmms):
        raise SystemExit(
            "At least one of --input_genomes, --input_orfs, or --input_hmms must be provided"
        )

    logger.info("Start annotation pipeline")

    run_annotate(
        args.input_genomes,
        args.input_orfs,
        args.input_hmms,
        args.output_directory,
        args.genomic_context_range,
        args.cpus,
    )


def run_profile_command(args):
    logger = logging.getLogger(__name__)

    r1 = args.read_1
    r2 = args.read_2
    s = args.single

    paired = r1 and r2

    if not s and not paired:
        raise SystemError(
            "At least either --read_1 and --read_2 must be provided or --single must be provided"
        )

    logger.info("Start mapping pipeline")

    run_profile(
        args.sample_id,
        r1,
        r2,
        s,
        args.output_directory,
        args.work_directory,
        args.cpus,
    )


def main():
    parser = get_parser()
    args = parser.parse_args()
    utils.configure_logging(args.verbose)
    args.func(args)


if __name__ == "__main__":
    main()



import subprocess
import logging
import sys
from pathlib import Path

from nfixplanet.constants import (
    NFIXPLANET_CACHE_DIR,
    REFERENCE_FASTA_URL,
    REFERENCE_FASTA_NAME,
    REFERENCE_INDEX_NAME,
)

logger = logging.getLogger(__name__)


# ---------------------------
# Generic helpers
# ---------------------------


def check_files_exist(paths: list[str]):
    for p in paths:
        if not Path(p).exists():
            logger.error(f"Missing file: {p}")
            sys.exit(1)


def run_command(cmd: list[str], error_msg: str):
    try:
        logger.info("Running: " + " ".join(cmd))
        subprocess.check_call(cmd)
    except subprocess.CalledProcessError as e:
        logger.error(f"{error_msg}: {e}")
        sys.exit(1)


# ---------------------------
# FASTP
# ---------------------------


def fastp_paired(
    r1_in: str,
    r2_in: str,
    r1_out: str,
    r2_out: str,
    unpaired: str,
    html: str,
    json: str,
    cpus: int,
):
    run_command(
        [
            "fastp",
            "--thread",
            str(cpus),
            "--in1",
            r1_in,
            "--in2",
            r2_in,
            "--out1",
            r1_out,
            "--out2",
            r2_out,
            "--unpaired1",
            unpaired,
            "--unpaired2",
            unpaired,
            "--html",
            html,
            "--json",
            json,
            "--length_required",
            "45",
            "--cut_front",
            "--cut_tail",
            "--cut_window_size",
            "4",
            "--cut_mean_quality",
            "20",
        ],
        f"fastp paired failed on {r1_in} {r2_in}",
    )


def fastp_single(
    s_in: str,
    s_out: str,
    html: str,
    json: str,
    cpus: int,
):
    run_command(
        [
            "fastp",
            "--thread",
            str(cpus),
            "--in1",
            s_in,
            "--out1",
            s_out,
            "--html",
            html,
            "--json",
            json,
            "--length_required",
            "45",
            "--cut_front",
            "--cut_tail",
            "--cut_window_size",
            "4",
            "--cut_mean_quality",
            "20",
        ],
        f"fastp single failed on {s_in}",
    )


# ---------------------------
# HOSTILE
# ---------------------------


def hostile_index_fetch():
    run_command(
        ["hostile", "index", "fetch"],
        "hostile index fetch failed",
    )


def hostile_clean_paired(r1: str, r2: str, out_dir: str, cpus: int):
    run_command(
        [
            "hostile",
            "clean",
            "--fastq1",
            r1,
            "--fastq2",
            r2,
            "--threads",
            str(cpus),
            "--output",
            out_dir,
        ],
        f"hostile clean paired failed on {r1} {r2}",
    )

    r1_out = f"{out_dir}/R1.clean_1.fastq.gz"
    r2_out = f"{out_dir}/R2.clean_2.fastq.gz"
    return r1_out, r2_out


def hostile_clean_single(s, out_dir, cpus: int):
    run_command(
        [
            "hostile",
            "clean",
            "--fastq1",
            s,
            "--threads",
            str(cpus),
            "--output",
            out_dir,
        ],
        f"hostile clean single failed on {s}",
    )

    return f"{out_dir}/RS.clean.fastq.gz"


# ---------------------------
# MINIMAP2
# ---------------------------


def fetch_reference_file():
    # make sure cache dir exists
    NFIXPLANET_CACHE_DIR.mkdir(parents=True, exist_ok=True)

    gz_path = f"{REFERENCE_FASTA_NAME}.gz"
    run_command(
        ["wget", "-O", gz_path, REFERENCE_FASTA_URL],
        "Failed to download reference file",
    )

    run_command(
        ["gunzip", gz_path],
        "Failed to unzip reference file",
    )


def minimap2_index(cpus: int):
    run_command(
        [
            "minimap2",
            "-d",
            str(REFERENCE_INDEX_NAME),
            "-t",
            str(cpus),
            str(REFERENCE_FASTA_NAME),
        ],
        "minimap2 index failed",
    )


def build_minimap_index(cpus: int):
    fetch_reference_file()
    # index
    minimap2_index(cpus)


# ---------------------------
# COVERM
# ---------------------------


def coverm_contig(
    r1: str | None,
    r2: str | None,
    single: str | None,
    reference_index: str,
    output_file: str,
    cpus: int,
):
    base_cmd = [
        "coverm",
        "contig",
        "-r",
        reference_index,
        "--threads",
        str(cpus),
        "--minimap2-reference-is-index",
        "--mapper",
        "minimap2-sr",
        "--methods",
        "covered_bases",
        "mean",
        "--output-file",
        output_file,
    ]

    if r1 and r2 and single:
        cmd = base_cmd + ["-1", r1, "-2", r2, "--single", single]
    elif r1 and r2:
        cmd = base_cmd + ["-1", r1, "-2", r2]
    elif single:
        cmd = base_cmd + ["--single", single]
    else:
        logger.error("No valid FASTQ inputs for coverm")
        sys.exit(1)

    run_command(cmd, "coverm failed")



import os
import pandas as pd
from pandas.testing import assert_frame_equal


REFERENCE_FOLDER = "/g/scb2/bork/ustick/nfixplanet_pipeline/01_annotation_pipeline/data/nfixplanet_test_full/QC_sequences"
GENERATED_FOLDER = "tmp_full"

def get_names() -> list[str]:
    names = []
    for f in os.listdir(GENERATED_FOLDER):
        if f.endswith(".tsv"):
            name = f.split(".")[0]
            names.append(name)
    return names


def compare_dfs(name: str):
    ref_file = f"{REFERENCE_FOLDER}/04_{name}_final.out"
    gen_file = f"{GENERATED_FOLDER}/{name}.tsv"

    ref_df = pd.read_csv(
        ref_file,
        sep="\t",
        header=None,
        names=[
            "query_name",
            "accession",
            "target_name",
            "full_accession",
            "full_evalue",
            "full_score",
            "full_bias",
            "domain_evalue",
            "domain_score",
            "domain_bias",
            "exp",
            "reg",
            "clu",
            "ov",
            "env",
            "dom",
            "rep",
            "inc",
        ],
    )
    ref_df = ref_df.sort_values(by=ref_df.columns[0]).reset_index(drop=True)
    gen_df = pd.read_csv(gen_file, sep="\t")
    gen_df = gen_df.sort_values(by=gen_df.columns[0]).reset_index(drop=True)

    # print(ref_df)
    # print(gen_df)
    try:
        assert_frame_equal(ref_df, gen_df, check_dtype=False)
    except Exception as e:
        print(f"failed: {name}")

        # # NOTE Rounding to ignore differences for now
        # # TODO: we have more results now, check if we are correct instead of the old results
        # float_cols = gen_df.select_dtypes('float').columns
        # gen_df[float_cols] = gen_df[float_cols].round(0).astype(int)
        # ref_df[float_cols] = ref_df[float_cols].round(0).astype(int)


        # gen_df.to_csv("gen_df.tsv", sep="\t", index=False)
        # ref_df.to_csv("ref_df.tsv", sep="\t", index=False)
        # # TODO: check 153;SAMN17048824;k141_265651_14 in hmm
        
        # raise e

print(f"Running on folder: {GENERATED_FOLDER}")
names = get_names()
for name in names:
    print(f"Checking {name}")
    compare_dfs(name)



import os
import logging
import shutil

logger = logging.getLogger(__name__)


def configure_logging(verbose: bool):
    level = logging.DEBUG if verbose else logging.INFO
    format = (
        "%(asctime)s : [%(levelname)7s] : %(name)s:%(lineno)s %(funcName)20s() : %(message)s"
        if verbose
        else "[%(asctime)s] %(levelname)s: %(message)s"
    )

    logging.basicConfig(
        format=format,
        datefmt="%H:%M:%S",
        level=level,
        force=True,
    )


def check_files_exist(files: list[str]) -> None:
    for file in files:
        if not os.path.isfile(file):
            raise FileNotFoundError(f"{file} does not exist")


def check_if_tool_exists(tool_name: str) -> bool:
    """Check if tool is available."""
    return shutil.which(tool_name) is not None


def check_external_tools(required_tools: list[str]):
    """Checks if tool dependencies are available."""
    for tool in required_tools:
        if not check_if_tool_exists(tool):
            logger.error(f"{tool} not found.")
            exit(1)

    logger.info("All required software found")
